// CForth LowPower project ***************************************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.h","cforthLP.c","cforthLP.h",
// and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32l031xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{

START:

    RCC->IOPENR |= RCC_IOPENR_GPIOAEN;  // Enable Port A clock
    RCC->IOPENR |= RCC_IOPENR_GPIOBEN;  // Enable Port B clock
    RCC->IOPENR |= RCC_IOPENR_GPIOCEN;  // Enable Port B clock

//	initGPIO();
    GPIOA->MODER = 0x00000000;      //primarily all input
    GPIOB->MODER = 0x00000000;      //primarily all input
    GPIOC->MODER = 0x00000000;      //primarily all input
    GPIOA->PUPDR = 0xAA8AAAAA;      //primarily all inputs pulldown-exceptPA10
    GPIOB->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown
    GPIOC->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown

//pin config repeated specifically:
	GPIOB->PUPDR &= ~(GPIO_PUPDR_PUPD0);	//pull-down off, wasts current
    GPIOB->MODER &= ~(GPIO_MODER_MODE0);
    GPIOB->MODER |= GPIO_MODER_MODE0_0 ;	//set PortB0 (LED) as output
    GPIOB->BSRR =  GPIO_BSRR_BS_0;			//PB0 (LED) ON
	gSpeed = 1;

	SetSysClock();							//1.048MHz MSI clock

    gLEDCOUNT = 0;
    gDELAYCOUNT = 0;
    gTIMEOUT = 0;
    gBaud = DEFAULT_BAUD;
    gBaudSet = 0;
	gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = DEFAULT_NUMBASE;
	gProject = 0;
	gLedBusy = 3;
	gfRandom = 314159265;		//seed for xorshift32 algorithm

    SysTick_Config(SystemCoreClock / 1000);	
								//for any SystemCoreClock
								//SystemCoreClock is derived from SYSCLOCK=MSI

    msDelay(50);

//init I/O state  *************************
    op_ABORT();					//reset default Forth system data
    popFlashPage205();			//loads only gBaud, gfNumBase and gProject

    msDelay(2);
 //init communication  ********************
    initUsart2();
	msDelay(2);

    sendTB(CRR);
    sendString(gTypeStr);
    sendTB(HARDWARE_VERSION + 0x30);
    sendTB('.');
    sendNS32(REVNUM);
    sendString(gCopyStr); 
	sendTB(CRR);

    popFlashProject();				//if (gProject==0) start empty project
									//if virgin or gProject==0 emptyForth();

    while(1) {
		quit();
		if (query() == 0) break;
		if (gfTibMax == 0) continue;
		eval();
    }
    msDelay(100);
    goto START;					//force cold reset 

}

// ##########################################################################

// Misc Low Level Routines
//************************

void SetSysClock(void)  // *************************************************
// *************************************************************************
{
__IO uint32_t StartUpCounter = 0, ClkStatus = 0;


//SYSCLK, HCLK, PCLK configuration  ----------------------------------------
//as precaution, disable PLL and HSE
	RCC->CR &= ~(RCC_CR_PLLON | RCC_CR_HSEON);
	RCC->CFGR &= ~(RCC_CFGR_PLLSRC | RCC_CFGR_PLLMUL);

	RCC->ICSCR &=  ~(0x00002000);	//bit 13 changes from MSIRANGE 5 to 4
									//now 1.048 MHz
	if (gSpeed == 4) RCC->ICSCR |= 0x00004000; //set bit14=4.194MHz
//Enable Prefetch Buffer //and set Flash Latency
	FLASH->ACR = FLASH_ACR_PRFTEN;	//=zero waitstates

	RCC->CFGR &= ~(RCC_CFGR_HPRE);		//clear all HPRE options
	RCC->CFGR |= RCC_CFGR_HPRE_DIV1;	//peripherals with max.speed
//PCLK = HCLK 
	RCC->CFGR |= (uint32_t)(RCC_CFGR_PPRE1_DIV1 | RCC_CFGR_PPRE2_DIV1);
	RCC->CCIPR = RCC_CCIPR_USART2SEL_0;	//SYSCLK selected for USART2

	SystemCoreClockUpdate();
/*
//possibly needed for SYSCLK freq test:
	RCC->CFGR &= (uint32_t)RCC_CFGR_MCOSEL;
	RCC->CFGR |= RCC_CFGR_MCOSEL_SYSCLK;
//set PA9 in alternate function mode
    GPIOA->MODER &= ~GPIO_MODER_MODE9;	//PA9 as output
    GPIOA->MODER |= GPIO_MODER_MODE9_0;
//MCO is AF0 on PA9
    GPIOA->AFR[1] &= 0xFFFFFF0F;        //PA9 is MCO
    GPIOA->AFR[1] |= 0x00000000;
*/

}

void msDelay(uint32_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
    gDELAYCOUNT = count;
    while(gDELAYCOUNT)  servSystem();		//unconditionally in this case
}


void wait5us(void)   // *****************************************************
// blocking delay ca 5us
// **************************************************************************
{
uint32_t countdown;
	countdown = 2;
    while (countdown)   {
        countdown-- ;
    }
}


void servSystem(void)   // ***************************************************
//not very time critical but permanently repeated functions for system control
//exclusively called in pullTermRxBuf
//this way it is regularily served when idle waiting for external input
//but watchdog resets when it hangs somewhere else
//slowed down by SysTick to 50 times/second
// ***************************************************************************
{
	if (gSYSCOUNT && (gSYSCOUNT < 0x80000000))  return;		
											//countdown in SysTick_Handler()

//else if if (gSYSCOUNT<=0 )20 Hz service interval:  *************************
    gSYSCOUNT = 50;         //restart 20Hz service interval

//LED control
	if (gLedBusy > 1)  {
		if (gLEDCOUNT) gLEDCOUNT--;
		if  (gLEDCOUNT==0) 	GPIOB->BSRR = GPIO_BSRR_BS_0;	//LED ON again
	}
//for non-blinking LED repeated here:
	else if (gLedBusy == 0)  GPIOB->BSRR = GPIO_BSRR_BR_0;//LED always OFF
    else  GPIOB->BSRR = GPIO_BSRR_BS_0;		//gLedBusy=1: LED now ON
}


void popFlashPage205(void)   // **********************************************
//read project independent system parameters
// **************************************************************************
{
uint32_t * flashPtr;
uint32_t test;

	emptyForth();			//primarily init defaults - implies gProject = 0;

	flashPtr = (uint32_t *)(PAGE205);
	test = *flashPtr;		//can't be=0 if project was saved in Flash
	flashPtr++;
	if (test == 0) return;	//defaults already init

	gBaud = *flashPtr;
	flashPtr++;

	gfNumBase = *flashPtr;
	flashPtr++;
	gfTempNumBase = gfNumBase;

	gProject = *flashPtr;
}


void pushFlashPage205(void)   //**********************************************
//save project independent system parameters permanently
// **************************************************************************
{
uint32_t *flashPtr;

//unlock flash
#define PEKEY1    0x89ABCDEF			//PEKEYR unlock key defined by STM
#define PEKEY2    0x02030405
#define PRGKEY1   0x8C9DAEBF			//PRGKEYR unlock key defined by STM
#define PRGKEY2   0x13141516

    while ((FLASH->SR & FLASH_SR_BSY) !=0);	//wait until no operation going on
    if ((FLASH->PECR & FLASH_PECR_PELOCK) != 0) {
            FLASH->PEKEYR = PEKEY1;
            FLASH->PEKEYR = PEKEY2;
    }
    while ((FLASH->SR & FLASH_SR_BSY) !=0);	//wait until no operation going on
    if ((FLASH->PECR & FLASH_PECR_PELOCK) == 0) {
		if ((FLASH->PECR & FLASH_PECR_PRGLOCK) != 0) {
				FLASH->PRGKEYR = PRGKEY1;
				FLASH->PRGKEYR = PRGKEY2;
		}
	}
    while ((FLASH->SR & FLASH_SR_BSY) !=0);	//wait until no operation going on

//erase Page205
    FLASH->PECR |= (FLASH_PECR_ERASE | FLASH_PECR_PROG);
    flashPtr = (uint32_t *)PAGE205;
	*flashPtr = 0;							//dummy for flash addressing
    while ((FLASH->SR & FLASH_SR_BSY) !=0);	//wait until no operation going on
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
	
//only write gBaud, gfNumBase and gProject,
//i.e.project independent parameters into page205

    flashPtr = (uint32_t *)PAGE205;
//	FLASH->PECR &= ~(FLASH_PECR_ERASE|FLASH_PECR_PROG);	//program single words
	FLASH->PECR &= ~(FLASH_PECR_ERASE);	//program single words
	*flashPtr = 0xFFFFFFFF;						//dummy for flash virgin test
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gBaud;
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gfNumBase;
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gProject;
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

//terminate
    FLASH->PECR &= ~(FLASH_PECR_ERASE | FLASH_PECR_PROG);
//lock flash
    FLASH->PECR |= FLASH_PECR_PELOCK;
    msDelay(10);
}


void popFlashProject(void)   //**********************************************
// **************************************************************************
{
uint32_t k, pwm;
uint32_t  *flashPtr;
uint32_t *sramPtr;

//emptyForth() already updated in popFlashPage205()(default system parameters)

	if (gProject)  {
		if (gProject == 1) flashPtr = (uint32_t *)PAGE206;
		else flashPtr = (uint32_t *)PAGE207;
// system and peripherial parameters
		k = *flashPtr;				//cannot be==0 if project saved in Flash
		flashPtr++;

		if (k== 0) {						//project Flash is virgin
			emptyForth();					//implies load gProject=0 instead
		}
		else   {
			gLedBusy = *flashPtr;
			flashPtr++;
			gfNumBase = *flashPtr;			//overrides setup from PAGE205!
			flashPtr++;
			gfTempNumBase = gfNumBase;
			gfAuto = *flashPtr;
			flashPtr++;
			gSpeed = *flashPtr;
			flashPtr++;

			gfSpiMode = *flashPtr;
			flashPtr++;
			gfSpiBits = *flashPtr;
			flashPtr++;
			gfSpiClock = *flashPtr;
			flashPtr++;
			gfBaseFreq = *flashPtr;
			flashPtr++;
			gfPWM1 = *flashPtr;
			flashPtr++;
#ifndef MOTOR
			gfPWM2 = *flashPtr;
#endif
			if (gfBaseFreq) {
				subFREQ();
				subPWM1();
#ifndef MOTOR
				if (gfPWM2 >= 0x7F) gfPWM2 = 0xFFFFFFFF;
				subPWM2();
#endif
			}
//first load UserToken Headers for project
//loading is possible over page limits
//gfUserHeader[] and gfTcode[] occupy a coherent data space in flash
			if (gProject == 1) flashPtr = (uint32_t *)PAGE208;
			else flashPtr = (uint32_t *)PAGE232;
			sramPtr = (uint32_t *)gfUserHeader;
			for (k = 0;k < (MAXUSER * 4);k++)  {	
										//384 words=1536 bytes=1.5pages
				*sramPtr = *flashPtr;
				sramPtr++;
				flashPtr++;
			}

//next load UserToken code for project
			sramPtr = (uint32_t *)gfTCode;
			for (k = 0;k < 0x180;k++)  {	//1536(0x600) bytes = 1.5 pages
				*sramPtr = *flashPtr;
				sramPtr++;
				flashPtr++;
			}
		}
	}

//from here if(gProject != 0) or not:
    wait5us();
    op_ABORT();					//reset project Forth system data
	if (gfAuto)  {
//emergency brake to avoid AUTO deadlock
		if ((GPIOA->IDR & 0x400) == 0) gfAuto = 0;	//if PA10 (RX input)is low
	}
    op_QUEST();
    if (gfAuto) {
        gfParsed = gfAuto - 1;	//-1 to distinguish from empty AUTOEXE == 0
        interpretUserToken();
    }
}


void pushFlashProject(void)   //*********************************************
// **************************************************************************
{
uint32_t p,k;
uint32_t  *flashPtr;
uint32_t *sramPtr;

	if ((gProject != 1) && (gProject != 2)) return;	//precaution

//unlock flash
#define PEKEY1    0x89ABCDEF			//PEKEYR unlock key defined by STM
#define PEKEY2    0x02030405
#define PRGKEY1   0x8C9DAEBF			//PRGKEYR unlock key defined by STM
#define PRGKEY2   0x13141516

    while ((FLASH->SR & FLASH_SR_BSY) !=0);	//wait until no operation going on
    if ((FLASH->PECR & FLASH_PECR_PELOCK) != 0) {
            FLASH->PEKEYR = PEKEY1;
            FLASH->PEKEYR = PEKEY2;
    }
    while ((FLASH->SR & FLASH_SR_BSY) !=0);	//wait until no operation going on
    if ((FLASH->PECR & FLASH_PECR_PELOCK) == 0) {
		if ((FLASH->PECR & FLASH_PECR_PRGLOCK) != 0) {
				FLASH->PRGKEYR = PRGKEY1;
				FLASH->PRGKEYR = PRGKEY2;
		}
	}
    while ((FLASH->SR & FLASH_SR_BSY) !=0);	//wait until no operation going on

//erase Page206 or 207
    FLASH->PECR |= (FLASH_PECR_ERASE | FLASH_PECR_PROG);

	if (gProject ==1) flashPtr = (uint32_t *)PAGE206;
	else flashPtr = (uint32_t *)PAGE207;
	*flashPtr = 0;							//dummy for flash addressing
    while ((FLASH->SR & FLASH_SR_BSY) !=0);	//wait until no operation going on
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;

//erase (Page208 to 231) or(Page232 to 255)
	for (k=0;k < 23;k++)  {
	if (gProject ==1) flashPtr = (uint32_t *)(PAGE208 + k*0x80);
	else flashPtr = (uint32_t *)(PAGE232 + k*0x80);
		*flashPtr = 0;						
		while ((FLASH->SR & FLASH_SR_BSY) !=0);
		if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
	}

//write system parameters  into page206 or 207
	FLASH->PECR &= ~(FLASH_PECR_ERASE);	//program single words
	if (gProject == 1) flashPtr = (uint32_t *)PAGE206;
	else flashPtr = (uint32_t *)PAGE207;
	FLASH->PECR &= ~(FLASH_PECR_PROG);		//program single words

	*flashPtr = 0xFFFFFFFF;					//used for Flash virgin test
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gLedBusy;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gfNumBase;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gfAuto;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr =gSpeed;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gfSpiMode;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gfSpiBits;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gfSpiClock;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gfBaseFreq;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

	*flashPtr = gfPWM1;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
#ifdef MOTOR
	*flashPtr = 0;
#else
	*flashPtr = gfPWM2;
#endif
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

//program UserToken headers (Page208 to 219) or (Page232 to 243)
	if (gProject ==1) flashPtr = (uint32_t *)PAGE208;
	else flashPtr = (uint32_t *)PAGE232;
	sramPtr = (uint32_t *)gfUserHeader;
	for (k = 0;k < (MAXUSER * 4);k++)  {	
		if ((k & 0xFF) == 0)  sendTB('.');
		*flashPtr = *sramPtr;
		flashPtr++;
		sramPtr++;
		while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
		if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	}

//program User code (Page220 to 231) or (Page244 to 255)
	if (gProject ==1) flashPtr = (uint32_t *)PAGE220;
	else flashPtr = (uint32_t *)PAGE244;
	sramPtr = (uint32_t *)gfTCode;
	for (k = 0;k < 0x180;k++)  {	//1536(0x600) bytes = 1.5 pages
		if ((k & 0xFF) == 0)  sendTB('.');
		*flashPtr = *sramPtr;
		flashPtr++;
		sramPtr++;
		while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
		if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	}
//terminate
    FLASH->PECR &= ~(FLASH_PECR_ERASE | FLASH_PECR_PROG | FLASH_PECR_FPRG);
//lock flash
    FLASH->PECR |= FLASH_PECR_PELOCK;
    msDelay(10);
}

//###########################################################################

//USART Support
//*************
void initUsart2(void) // ****************************************************
// USART2-TxD is PortB,6
// USART2-RxD is PortA,10 in any case
// USART2 clock is always derived from SYSCLOCK= 1.048 or 4.194 MHz
// **************************************************************************
{
    RCC->APB1ENR |=  RCC_APB1ENR_USART2EN ;	// USART2 clock enable

//set PA10 in alternate function mode
    GPIOA->MODER &= ~GPIO_MODER_MODE10_0 ;
    GPIOA->MODER |= GPIO_MODER_MODE10_1;
//set PB6 in alternate function mode
    GPIOB->MODER &= ~GPIO_MODER_MODE6_0 ;
    GPIOB->MODER |= GPIO_MODER_MODE6_1;
//USART2 is AF4 on PortA
    GPIOA->AFR[1] &= 0xFFFFF0FF;        //PA10 is RX
    GPIOA->AFR[1] |= 0x00000400;
//and AF0 on PortB
    GPIOB->AFR[0] &= 0xF0FFFFFF;        //PB6 is TX
    GPIOB->AFR[0] |= 0x00000000;
    NVIC_DisableIRQ(USART2_IRQn);
	subInitUsart2();
 //   NVIC_EnableIRQ(USART2_IRQn);
//	msDelay(2);
    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;
    NVIC_EnableIRQ(USART2_IRQn);
	msDelay(2);
}


void subInitUsart2(void) // **************************************************
// separated for BaudRate change without re-init buffer
// USART2 clock is always derived from SYSCLOCK= 1.048 or 4.194 MHz
// **************************************************************************
{
uint16_t baudset;

    USART2->CR1 &= ~(USART_CR1_UE);	//disable USART - else no Baudrate change
	if (gSpeed == 1)  {
		switch(gBaud) {
		case 96:
			baudset = 109;			//9600Baud:109.17 at 1.048MHz SYSCL
			break;
		case 38:
			baudset = 27;			//38400Baud: 27.29 at 1.048MHz SYSCLK
			break;
		case 19:
			baudset = 53;			//19200Baud 53.46 at 1.048MHz SYSCLK
			break;
		case 57:
		default:
			baudset = 18;			//57600Baud: 18.19 at 1.048MHz SYSCLK
			break;
		}
	}
	else  {
		switch(gBaud) {
		case 96:
			baudset = 437;			//9600Baud:436.9 at 4.194MHz SYSCL
			break;
		case 38:
			baudset = 109;			//38400Baud: 109.2 at 4.194MHz SYSCLK
			break;
		case 19:
			baudset = 218;			//19200Baud 218.4 at 4.194MHz SYSCLK
			break;
		case 57:
		default:
			baudset = 73;			//57600Baud: 72.8 at 4.194MHz SYSCLK
			break;
		}
	}
    USART2->BRR = baudset;
	USART2->CR2 = USART_CR2_TXINV;	//1 Stopbit, TX level inverted
    USART2->CR3 = 0;				//default:no Handshake
    USART2->CR1 = USART_CR1_RXNEIE | USART_CR1_TE | USART_CR1_RE;
									//8Bit, no Parity, RX Int
    USART2->CR1 |= USART_CR1_UE;	//enable USART
    gBaudSet = 1;
}


void sendTB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    while(!(USART2->ISR&USART_ISR_TXE));
    USART2->TDR = data;
	if (gLedBusy == 3) {
		GPIOB->BSRR = GPIO_BSRR_BR_0;	// clear bit PB0,LED OFF
		gLEDCOUNT = 4;					// ca.30 milliseconds
	}
}


void sendNibble(uint8_t data) // ********************************************
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
	sendTB(asciiNibble);
}


void sendString(const uint8_t * str) // *************************************
// **************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendTB(byte);
}


void sendX8(uint32_t data) // *********************************************
// send bits 0...7 in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data & 0xF0) >> 4;
    temp = temp  +0x30;
    if (temp > 0x39) temp+= 7;
    sendTB(temp);
    temp= data & 0x0F;
    temp = temp + 0x30;
    if (temp > 0x39) temp += 7;
    sendTB(temp);
    sendTB(0x20);
}


void sendX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// Leading zeroes are suppressed
// **************************************************************************
{
uint8_t nibble;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
    sendTB(0x20);

}


void sendNX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble, noLead;
    noLead = 0;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    if (nibble)  {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
     if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
}


void send0X32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format WITH leading 0x.Leading zeroes suppressed
// **************************************************************************
{
//	if (data >= 10)  {			//modified 11Jan24
		sendTB('0');
		sendTB('x');
//	}
	sendNX32(data);
}


void sendS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendNS32(data);
    sendTB(0x20);
}


 void sendLS32(uint32_t data) // ********************************************
//send signed decimal number string. Leading zeroes are suppressed
//leading SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendTB(0x20);
    sendNS32(data);
}


void sendNS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//no leading and trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendTB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendTB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendTB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendTB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendTB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendTB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendTB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendTB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendTB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendTB((uint8_t)quotient + 0x30);
ONES:
    sendTB((uint8_t)data + 0x30) ;
}


uint32_t pullTermRxBuf(void) // ********************************************
// pull 1 byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

	servSystem();		//serve watchdog etc.

	if (gUSART_RXTAIL!=gUSART_RXHEAD)  {
		recv= (uint32_t)((*gUSART_RXTAIL++ & 0x000000FF)|0x00008000);
		if (gUSART_RXTAIL==gUSART_RXBUFTOP)  gUSART_RXTAIL=gUSART_RXBUF;

		if (gLedBusy > 1) {
			GPIOB->BSRR = GPIO_BSRR_BR_0;	// clear bit PB0,LED OFF
			gLEDCOUNT = 4;					//ca.300 milliseconds
		}
		return recv;
	}
	else return 0;
}


// Interrupt Handlers
//*******************

void SysTick_Handler(void)  // ***********************************************
// ***************************************************************************
{
	if (gSYSCOUNT) gSYSCOUNT--;			//controls servSystem()
	if (gDELAYCOUNT) gDELAYCOUNT--;
	if (gTIMEOUT) gTIMEOUT--;			//to be used in any subroutine
	if (gfMSCOUNT) gfMSCOUNT --;		//used by MS, DS
	if (gfTIXCOUNT) gfTIXCOUNT --;		//used by  TIX, WAIT
#ifdef MOTOR
	if (gfSTEPCOUNT > 0) gfSTEPCOUNT--;		//stepper timing
#endif
	gfRandom ^= gfRandom << 13;				//xorshift32 algorithm/white noise
	gfRandom ^= gfRandom >> 17;
	gfRandom ^= gfRandom << 5;
}

void USART2_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t test;

	if (USART2->ISR & USART_ISR_RXNE)  {
		test = USART2->RDR & 0x00FF;
		if (test == CTRL_B) gfTermInput = CTRL_B;
		if (test != CTRL_D)  {
			*gUSART_RXHEAD = (uint8_t)test;
			gUSART_RXHEAD++;
			if (gUSART_RXHEAD==gUSART_RXBUFTOP)  gUSART_RXHEAD=gUSART_RXBUF;
		}
        else  op_DEBUG();
	}
}

// #########################################################################
